/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/. */





    /// https://www.w3.org/TR/SVG2/painting.html#MarkerShorthand
    pub mod marker {
        use cssparser::Parser;
        use crate::parser::ParserContext;
        use crate::properties::{PropertyDeclaration, SourcePropertyDeclaration, MaybeBoxed, longhands};
        #[allow(unused_imports)]
        use selectors::parser::SelectorParseErrorKind;
        #[allow(unused_imports)]
        use std::fmt::{self, Write};
        #[allow(unused_imports)]
        use style_traits::{ParseError, StyleParseErrorKind};
        #[allow(unused_imports)]
        use style_traits::{CssWriter, KeywordsCollectFn, SpecifiedValueInfo, ToCss};

        #[derive(SpecifiedValueInfo)]
        pub struct Longhands {
                pub marker_start:
                    longhands::marker_start::SpecifiedValue
                    ,
                pub marker_end:
                    longhands::marker_end::SpecifiedValue
                    ,
                pub marker_mid:
                    longhands::marker_mid::SpecifiedValue
                    ,
        }

        /// Represents a serializable set of all of the longhand properties that
        /// correspond to a shorthand.
        pub struct LonghandsToSerialize<'a> {
                pub marker_start:
                    &'a longhands::marker_start::SpecifiedValue,
                pub marker_end:
                    &'a longhands::marker_end::SpecifiedValue,
                pub marker_mid:
                    &'a longhands::marker_mid::SpecifiedValue,
        }

        impl<'a> LonghandsToSerialize<'a> {
            /// Tries to get a serializable set of longhands given a set of
            /// property declarations.
            pub fn from_iter<I>(iter: I) -> Result<Self, ()>
            where
                I: Iterator<Item=&'a PropertyDeclaration>,
            {
                // Define all of the expected variables that correspond to the shorthand
                    let mut marker_start = None;
                    let mut marker_end = None;
                    let mut marker_mid = None;

                // Attempt to assign the incoming declarations to the expected variables
                for longhand in iter {
                    match *longhand {
                            PropertyDeclaration::MarkerStart(ref value) => {
                                marker_start = Some(value)
                            },
                            PropertyDeclaration::MarkerEnd(ref value) => {
                                marker_end = Some(value)
                            },
                            PropertyDeclaration::MarkerMid(ref value) => {
                                marker_mid = Some(value)
                            },
                        _ => {}
                    };
                }

                // If any of the expected variables are missing, return an error
                match (
                        marker_start,
                        marker_end,
                        marker_mid,
                ) {

                    (
                        Some(marker_start),
                        Some(marker_end),
                        Some(marker_mid),
                    ) =>
                    Ok(LonghandsToSerialize {
                            marker_start,
                            marker_end,
                            marker_mid,
                    }),
                    _ => Err(())
                }
            }
        }

        /// Parse the given shorthand and fill the result into the
        /// `declarations` vector.
        pub fn parse_into<'i, 't>(
            declarations: &mut SourcePropertyDeclaration,
            context: &ParserContext,
            input: &mut Parser<'i, 't>,
        ) -> Result<(), ParseError<'i>> {
            #[allow(unused_imports)]
            use crate::properties::{NonCustomPropertyId, LonghandId};
            input.parse_entirely(|input| parse_value(context, input)).map(|longhands| {
                    declarations.push(PropertyDeclaration::MarkerStart(
                        longhands.marker_start
                    ));
                    declarations.push(PropertyDeclaration::MarkerEnd(
                        longhands.marker_end
                    ));
                    declarations.push(PropertyDeclaration::MarkerMid(
                        longhands.marker_mid
                    ));
            })
        }

        
    use crate::values::specified::url::UrlOrNone;

    pub fn parse_value<'i, 't>(
        context: &ParserContext,
        input: &mut Parser<'i, 't>,
    ) -> Result<Longhands, ParseError<'i>> {
        use crate::parser::Parse;
        let url = UrlOrNone::parse(context, input)?;

        Ok(expanded! {
            marker_start: url.clone(),
            marker_mid: url.clone(),
            marker_end: url,
        })
    }

    impl<'a> ToCss for LonghandsToSerialize<'a>  {
        fn to_css<W>(&self, dest: &mut CssWriter<W>) -> fmt::Result where W: fmt::Write {
            if self.marker_start == self.marker_mid && self.marker_mid == self.marker_end {
                self.marker_start.to_css(dest)
            } else {
                Ok(())
            }
        }
    }

    }

