/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/. */





    /// https://drafts.csswg.org/css-ui/#propdef-outline
    pub mod outline {
        use cssparser::Parser;
        use crate::parser::ParserContext;
        use crate::properties::{PropertyDeclaration, SourcePropertyDeclaration, MaybeBoxed, longhands};
        #[allow(unused_imports)]
        use selectors::parser::SelectorParseErrorKind;
        #[allow(unused_imports)]
        use std::fmt::{self, Write};
        #[allow(unused_imports)]
        use style_traits::{ParseError, StyleParseErrorKind};
        #[allow(unused_imports)]
        use style_traits::{CssWriter, KeywordsCollectFn, SpecifiedValueInfo, ToCss};

        #[derive(SpecifiedValueInfo)]
        pub struct Longhands {
                pub outline_color:
                    longhands::outline_color::SpecifiedValue
                    ,
                pub outline_style:
                    longhands::outline_style::SpecifiedValue
                    ,
                pub outline_width:
                    longhands::outline_width::SpecifiedValue
                    ,
        }

        /// Represents a serializable set of all of the longhand properties that
        /// correspond to a shorthand.
        #[derive(ToCss)]
        pub struct LonghandsToSerialize<'a> {
                pub outline_color:
                    &'a longhands::outline_color::SpecifiedValue,
                pub outline_style:
                    &'a longhands::outline_style::SpecifiedValue,
                pub outline_width:
                    &'a longhands::outline_width::SpecifiedValue,
        }

        impl<'a> LonghandsToSerialize<'a> {
            /// Tries to get a serializable set of longhands given a set of
            /// property declarations.
            pub fn from_iter<I>(iter: I) -> Result<Self, ()>
            where
                I: Iterator<Item=&'a PropertyDeclaration>,
            {
                // Define all of the expected variables that correspond to the shorthand
                    let mut outline_color = None;
                    let mut outline_style = None;
                    let mut outline_width = None;

                // Attempt to assign the incoming declarations to the expected variables
                for longhand in iter {
                    match *longhand {
                            PropertyDeclaration::OutlineColor(ref value) => {
                                outline_color = Some(value)
                            },
                            PropertyDeclaration::OutlineStyle(ref value) => {
                                outline_style = Some(value)
                            },
                            PropertyDeclaration::OutlineWidth(ref value) => {
                                outline_width = Some(value)
                            },
                        _ => {}
                    };
                }

                // If any of the expected variables are missing, return an error
                match (
                        outline_color,
                        outline_style,
                        outline_width,
                ) {

                    (
                        Some(outline_color),
                        Some(outline_style),
                        Some(outline_width),
                    ) =>
                    Ok(LonghandsToSerialize {
                            outline_color,
                            outline_style,
                            outline_width,
                    }),
                    _ => Err(())
                }
            }
        }

        /// Parse the given shorthand and fill the result into the
        /// `declarations` vector.
        pub fn parse_into<'i, 't>(
            declarations: &mut SourcePropertyDeclaration,
            context: &ParserContext,
            input: &mut Parser<'i, 't>,
        ) -> Result<(), ParseError<'i>> {
            #[allow(unused_imports)]
            use crate::properties::{NonCustomPropertyId, LonghandId};
            input.parse_entirely(|input| parse_value(context, input)).map(|longhands| {
                    declarations.push(PropertyDeclaration::OutlineColor(
                        longhands.outline_color
                    ));
                    declarations.push(PropertyDeclaration::OutlineStyle(
                        longhands.outline_style
                    ));
                    declarations.push(PropertyDeclaration::OutlineWidth(
                        longhands.outline_width
                    ));
            })
        }

        
    use crate::properties::longhands::{outline_color, outline_width, outline_style};
    use crate::values::specified;
    use crate::parser::Parse;

    pub fn parse_value<'i, 't>(
        context: &ParserContext,
        input: &mut Parser<'i, 't>,
    ) -> Result<Longhands, ParseError<'i>> {
        let _unused = context;
        let mut color = None;
        let mut style = None;
        let mut width = None;
        let mut any = false;
        loop {
            if color.is_none() {
                if let Ok(value) = input.try(|i| specified::Color::parse(context, i)) {
                    color = Some(value);
                    any = true;
                    continue
                }
            }
            if style.is_none() {
                if let Ok(value) = input.try(|input| outline_style::parse(context, input)) {
                    style = Some(value);
                    any = true;
                    continue
                }
            }
            if width.is_none() {
                if let Ok(value) = input.try(|input| outline_width::parse(context, input)) {
                    width = Some(value);
                    any = true;
                    continue
                }
            }
            break
        }
        if any {
            Ok(expanded! {
                outline_color: unwrap_or_initial!(outline_color, color),
                outline_style: unwrap_or_initial!(outline_style, style),
                outline_width: unwrap_or_initial!(outline_width, width),
            })
        } else {
            Err(input.new_custom_error(StyleParseErrorKind::UnspecifiedError))
        }
    }

    }


// The -moz-outline-radius shorthand is non-standard and not on a standards track.


    /// Nonstandard (https://developer.mozilla.org/en-US/docs/Web/CSS/-moz-outline-radius)
    pub mod _moz_outline_radius {
        use cssparser::Parser;
        use crate::parser::ParserContext;
        use crate::properties::{PropertyDeclaration, SourcePropertyDeclaration, MaybeBoxed, longhands};
        #[allow(unused_imports)]
        use selectors::parser::SelectorParseErrorKind;
        #[allow(unused_imports)]
        use std::fmt::{self, Write};
        #[allow(unused_imports)]
        use style_traits::{ParseError, StyleParseErrorKind};
        #[allow(unused_imports)]
        use style_traits::{CssWriter, KeywordsCollectFn, SpecifiedValueInfo, ToCss};

        #[derive(SpecifiedValueInfo)]
        pub struct Longhands {
                pub _moz_outline_radius_topleft:
                        Box<
                    longhands::_moz_outline_radius_topleft::SpecifiedValue
                        >
                    ,
                pub _moz_outline_radius_topright:
                        Box<
                    longhands::_moz_outline_radius_topright::SpecifiedValue
                        >
                    ,
                pub _moz_outline_radius_bottomright:
                        Box<
                    longhands::_moz_outline_radius_bottomright::SpecifiedValue
                        >
                    ,
                pub _moz_outline_radius_bottomleft:
                        Box<
                    longhands::_moz_outline_radius_bottomleft::SpecifiedValue
                        >
                    ,
        }

        /// Represents a serializable set of all of the longhand properties that
        /// correspond to a shorthand.
        pub struct LonghandsToSerialize<'a> {
                pub _moz_outline_radius_topleft:
                    &'a longhands::_moz_outline_radius_topleft::SpecifiedValue,
                pub _moz_outline_radius_topright:
                    &'a longhands::_moz_outline_radius_topright::SpecifiedValue,
                pub _moz_outline_radius_bottomright:
                    &'a longhands::_moz_outline_radius_bottomright::SpecifiedValue,
                pub _moz_outline_radius_bottomleft:
                    &'a longhands::_moz_outline_radius_bottomleft::SpecifiedValue,
        }

        impl<'a> LonghandsToSerialize<'a> {
            /// Tries to get a serializable set of longhands given a set of
            /// property declarations.
            pub fn from_iter<I>(iter: I) -> Result<Self, ()>
            where
                I: Iterator<Item=&'a PropertyDeclaration>,
            {
                // Define all of the expected variables that correspond to the shorthand
                    let mut _moz_outline_radius_topleft = None;
                    let mut _moz_outline_radius_topright = None;
                    let mut _moz_outline_radius_bottomright = None;
                    let mut _moz_outline_radius_bottomleft = None;

                // Attempt to assign the incoming declarations to the expected variables
                for longhand in iter {
                    match *longhand {
                            PropertyDeclaration::MozOutlineRadiusTopleft(ref value) => {
                                _moz_outline_radius_topleft = Some(value)
                            },
                            PropertyDeclaration::MozOutlineRadiusTopright(ref value) => {
                                _moz_outline_radius_topright = Some(value)
                            },
                            PropertyDeclaration::MozOutlineRadiusBottomright(ref value) => {
                                _moz_outline_radius_bottomright = Some(value)
                            },
                            PropertyDeclaration::MozOutlineRadiusBottomleft(ref value) => {
                                _moz_outline_radius_bottomleft = Some(value)
                            },
                        _ => {}
                    };
                }

                // If any of the expected variables are missing, return an error
                match (
                        _moz_outline_radius_topleft,
                        _moz_outline_radius_topright,
                        _moz_outline_radius_bottomright,
                        _moz_outline_radius_bottomleft,
                ) {

                    (
                        Some(_moz_outline_radius_topleft),
                        Some(_moz_outline_radius_topright),
                        Some(_moz_outline_radius_bottomright),
                        Some(_moz_outline_radius_bottomleft),
                    ) =>
                    Ok(LonghandsToSerialize {
                            _moz_outline_radius_topleft,
                            _moz_outline_radius_topright,
                            _moz_outline_radius_bottomright,
                            _moz_outline_radius_bottomleft,
                    }),
                    _ => Err(())
                }
            }
        }

        /// Parse the given shorthand and fill the result into the
        /// `declarations` vector.
        pub fn parse_into<'i, 't>(
            declarations: &mut SourcePropertyDeclaration,
            context: &ParserContext,
            input: &mut Parser<'i, 't>,
        ) -> Result<(), ParseError<'i>> {
            #[allow(unused_imports)]
            use crate::properties::{NonCustomPropertyId, LonghandId};
            input.parse_entirely(|input| parse_value(context, input)).map(|longhands| {
                    declarations.push(PropertyDeclaration::MozOutlineRadiusTopleft(
                        longhands._moz_outline_radius_topleft
                    ));
                    declarations.push(PropertyDeclaration::MozOutlineRadiusTopright(
                        longhands._moz_outline_radius_topright
                    ));
                    declarations.push(PropertyDeclaration::MozOutlineRadiusBottomright(
                        longhands._moz_outline_radius_bottomright
                    ));
                    declarations.push(PropertyDeclaration::MozOutlineRadiusBottomleft(
                        longhands._moz_outline_radius_bottomleft
                    ));
            })
        }

        
    use crate::values::generics::rect::Rect;
    use crate::values::specified::border::BorderRadius;
    use crate::parser::Parse;

    pub fn parse_value<'i, 't>(
        context: &ParserContext,
        input: &mut Parser<'i, 't>,
    ) -> Result<Longhands, ParseError<'i>> {
        let radii = BorderRadius::parse(context, input)?;
        Ok(expanded! {
            _moz_outline_radius_topleft: radii.top_left,
            _moz_outline_radius_topright: radii.top_right,
            _moz_outline_radius_bottomright: radii.bottom_right,
            _moz_outline_radius_bottomleft: radii.bottom_left,
        })
    }

    impl<'a> ToCss for LonghandsToSerialize<'a>  {
        fn to_css<W>(&self, dest: &mut CssWriter<W>) -> fmt::Result where W: fmt::Write {
            use crate::values::generics::border::BorderCornerRadius;

            let LonghandsToSerialize {
                _moz_outline_radius_topleft: &BorderCornerRadius(ref tl),
                _moz_outline_radius_topright: &BorderCornerRadius(ref tr),
                _moz_outline_radius_bottomright: &BorderCornerRadius(ref br),
                _moz_outline_radius_bottomleft: &BorderCornerRadius(ref bl),
            } = *self;

            let widths = Rect::new(tl.width(), tr.width(), br.width(), bl.width());
            let heights = Rect::new(tl.height(), tr.height(), br.height(), bl.height());

            BorderRadius::serialize_rects(widths, heights, dest)
        }
    }

    }

