/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/. */





    /// https://drafts.fxtf.org/css-masking/#propdef-mask
    pub mod mask {
        use cssparser::Parser;
        use crate::parser::ParserContext;
        use crate::properties::{PropertyDeclaration, SourcePropertyDeclaration, MaybeBoxed, longhands};
        #[allow(unused_imports)]
        use selectors::parser::SelectorParseErrorKind;
        #[allow(unused_imports)]
        use std::fmt::{self, Write};
        #[allow(unused_imports)]
        use style_traits::{ParseError, StyleParseErrorKind};
        #[allow(unused_imports)]
        use style_traits::{CssWriter, KeywordsCollectFn, SpecifiedValueInfo, ToCss};

        #[derive(SpecifiedValueInfo)]
        pub struct Longhands {
                pub mask_mode:
                    longhands::mask_mode::SpecifiedValue
                    ,
                pub mask_repeat:
                    longhands::mask_repeat::SpecifiedValue
                    ,
                pub mask_clip:
                    longhands::mask_clip::SpecifiedValue
                    ,
                pub mask_origin:
                    longhands::mask_origin::SpecifiedValue
                    ,
                pub mask_composite:
                    longhands::mask_composite::SpecifiedValue
                    ,
                pub mask_position_x:
                    longhands::mask_position_x::SpecifiedValue
                    ,
                pub mask_position_y:
                    longhands::mask_position_y::SpecifiedValue
                    ,
                pub mask_size:
                    longhands::mask_size::SpecifiedValue
                    ,
                pub mask_image:
                    longhands::mask_image::SpecifiedValue
                    ,
        }

        /// Represents a serializable set of all of the longhand properties that
        /// correspond to a shorthand.
        pub struct LonghandsToSerialize<'a> {
                pub mask_mode:
                    &'a longhands::mask_mode::SpecifiedValue,
                pub mask_repeat:
                    &'a longhands::mask_repeat::SpecifiedValue,
                pub mask_clip:
                    &'a longhands::mask_clip::SpecifiedValue,
                pub mask_origin:
                    &'a longhands::mask_origin::SpecifiedValue,
                pub mask_composite:
                    &'a longhands::mask_composite::SpecifiedValue,
                pub mask_position_x:
                    &'a longhands::mask_position_x::SpecifiedValue,
                pub mask_position_y:
                    &'a longhands::mask_position_y::SpecifiedValue,
                pub mask_size:
                    &'a longhands::mask_size::SpecifiedValue,
                pub mask_image:
                    &'a longhands::mask_image::SpecifiedValue,
        }

        impl<'a> LonghandsToSerialize<'a> {
            /// Tries to get a serializable set of longhands given a set of
            /// property declarations.
            pub fn from_iter<I>(iter: I) -> Result<Self, ()>
            where
                I: Iterator<Item=&'a PropertyDeclaration>,
            {
                // Define all of the expected variables that correspond to the shorthand
                    let mut mask_mode = None;
                    let mut mask_repeat = None;
                    let mut mask_clip = None;
                    let mut mask_origin = None;
                    let mut mask_composite = None;
                    let mut mask_position_x = None;
                    let mut mask_position_y = None;
                    let mut mask_size = None;
                    let mut mask_image = None;

                // Attempt to assign the incoming declarations to the expected variables
                for longhand in iter {
                    match *longhand {
                            PropertyDeclaration::MaskMode(ref value) => {
                                mask_mode = Some(value)
                            },
                            PropertyDeclaration::MaskRepeat(ref value) => {
                                mask_repeat = Some(value)
                            },
                            PropertyDeclaration::MaskClip(ref value) => {
                                mask_clip = Some(value)
                            },
                            PropertyDeclaration::MaskOrigin(ref value) => {
                                mask_origin = Some(value)
                            },
                            PropertyDeclaration::MaskComposite(ref value) => {
                                mask_composite = Some(value)
                            },
                            PropertyDeclaration::MaskPositionX(ref value) => {
                                mask_position_x = Some(value)
                            },
                            PropertyDeclaration::MaskPositionY(ref value) => {
                                mask_position_y = Some(value)
                            },
                            PropertyDeclaration::MaskSize(ref value) => {
                                mask_size = Some(value)
                            },
                            PropertyDeclaration::MaskImage(ref value) => {
                                mask_image = Some(value)
                            },
                        _ => {}
                    };
                }

                // If any of the expected variables are missing, return an error
                match (
                        mask_mode,
                        mask_repeat,
                        mask_clip,
                        mask_origin,
                        mask_composite,
                        mask_position_x,
                        mask_position_y,
                        mask_size,
                        mask_image,
                ) {

                    (
                        Some(mask_mode),
                        Some(mask_repeat),
                        Some(mask_clip),
                        Some(mask_origin),
                        Some(mask_composite),
                        Some(mask_position_x),
                        Some(mask_position_y),
                        Some(mask_size),
                        Some(mask_image),
                    ) =>
                    Ok(LonghandsToSerialize {
                            mask_mode,
                            mask_repeat,
                            mask_clip,
                            mask_origin,
                            mask_composite,
                            mask_position_x,
                            mask_position_y,
                            mask_size,
                            mask_image,
                    }),
                    _ => Err(())
                }
            }
        }

        /// Parse the given shorthand and fill the result into the
        /// `declarations` vector.
        pub fn parse_into<'i, 't>(
            declarations: &mut SourcePropertyDeclaration,
            context: &ParserContext,
            input: &mut Parser<'i, 't>,
        ) -> Result<(), ParseError<'i>> {
            #[allow(unused_imports)]
            use crate::properties::{NonCustomPropertyId, LonghandId};
            input.parse_entirely(|input| parse_value(context, input)).map(|longhands| {
                    declarations.push(PropertyDeclaration::MaskMode(
                        longhands.mask_mode
                    ));
                    declarations.push(PropertyDeclaration::MaskRepeat(
                        longhands.mask_repeat
                    ));
                    declarations.push(PropertyDeclaration::MaskClip(
                        longhands.mask_clip
                    ));
                    declarations.push(PropertyDeclaration::MaskOrigin(
                        longhands.mask_origin
                    ));
                    declarations.push(PropertyDeclaration::MaskComposite(
                        longhands.mask_composite
                    ));
                    declarations.push(PropertyDeclaration::MaskPositionX(
                        longhands.mask_position_x
                    ));
                    declarations.push(PropertyDeclaration::MaskPositionY(
                        longhands.mask_position_y
                    ));
                    declarations.push(PropertyDeclaration::MaskSize(
                        longhands.mask_size
                    ));
                    declarations.push(PropertyDeclaration::MaskImage(
                        longhands.mask_image
                    ));
            })
        }

        
    use crate::properties::longhands::{mask_mode, mask_repeat, mask_clip, mask_origin, mask_composite, mask_position_x,
                                mask_position_y};
    use crate::properties::longhands::{mask_size, mask_image};
    use crate::values::specified::{Position, PositionComponent};
    use crate::parser::Parse;

    // FIXME(emilio): These two mask types should be the same!
    impl From<mask_origin::single_value::SpecifiedValue> for mask_clip::single_value::SpecifiedValue {
        fn from(origin: mask_origin::single_value::SpecifiedValue) -> mask_clip::single_value::SpecifiedValue {
            match origin {
                mask_origin::single_value::SpecifiedValue::ContentBox =>
                    mask_clip::single_value::SpecifiedValue::ContentBox,
                mask_origin::single_value::SpecifiedValue::PaddingBox =>
                    mask_clip::single_value::SpecifiedValue::PaddingBox ,
                mask_origin::single_value::SpecifiedValue::BorderBox =>
                    mask_clip::single_value::SpecifiedValue::BorderBox,
                mask_origin::single_value::SpecifiedValue::FillBox =>
                    mask_clip::single_value::SpecifiedValue::FillBox ,
                mask_origin::single_value::SpecifiedValue::StrokeBox =>
                    mask_clip::single_value::SpecifiedValue::StrokeBox,
                mask_origin::single_value::SpecifiedValue::ViewBox=>
                    mask_clip::single_value::SpecifiedValue::ViewBox,
            }
        }
    }

    pub fn parse_value<'i, 't>(
        context: &ParserContext,
        input: &mut Parser<'i, 't>,
    ) -> Result<Longhands, ParseError<'i>> {
        // Vec grows from 0 to 4 by default on first push().  So allocate with
        // capacity 1, so in the common case of only one item we don't way
        // overallocate, then shrink.  Note that we always push at least one
        // item if parsing succeeds.
        let mut mask_image = Vec::with_capacity(1);
        // Vec grows from 0 to 4 by default on first push().  So allocate with
        // capacity 1, so in the common case of only one item we don't way
        // overallocate, then shrink.  Note that we always push at least one
        // item if parsing succeeds.
        let mut mask_mode = Vec::with_capacity(1);
        // Vec grows from 0 to 4 by default on first push().  So allocate with
        // capacity 1, so in the common case of only one item we don't way
        // overallocate, then shrink.  Note that we always push at least one
        // item if parsing succeeds.
        let mut mask_position_x = Vec::with_capacity(1);
        // Vec grows from 0 to 4 by default on first push().  So allocate with
        // capacity 1, so in the common case of only one item we don't way
        // overallocate, then shrink.  Note that we always push at least one
        // item if parsing succeeds.
        let mut mask_position_y = Vec::with_capacity(1);
        // Vec grows from 0 to 4 by default on first push().  So allocate with
        // capacity 1, so in the common case of only one item we don't way
        // overallocate, then shrink.  Note that we always push at least one
        // item if parsing succeeds.
        let mut mask_size = Vec::with_capacity(1);
        // Vec grows from 0 to 4 by default on first push().  So allocate with
        // capacity 1, so in the common case of only one item we don't way
        // overallocate, then shrink.  Note that we always push at least one
        // item if parsing succeeds.
        let mut mask_repeat = Vec::with_capacity(1);
        // Vec grows from 0 to 4 by default on first push().  So allocate with
        // capacity 1, so in the common case of only one item we don't way
        // overallocate, then shrink.  Note that we always push at least one
        // item if parsing succeeds.
        let mut mask_origin = Vec::with_capacity(1);
        // Vec grows from 0 to 4 by default on first push().  So allocate with
        // capacity 1, so in the common case of only one item we don't way
        // overallocate, then shrink.  Note that we always push at least one
        // item if parsing succeeds.
        let mut mask_clip = Vec::with_capacity(1);
        // Vec grows from 0 to 4 by default on first push().  So allocate with
        // capacity 1, so in the common case of only one item we don't way
        // overallocate, then shrink.  Note that we always push at least one
        // item if parsing succeeds.
        let mut mask_composite = Vec::with_capacity(1);

        input.parse_comma_separated(|input| {
                let mut image = None;
                let mut mode = None;
                let mut position = None;
                let mut size = None;
                let mut repeat = None;
                let mut origin = None;
                let mut clip = None;
                let mut composite = None;
            loop {
                if image.is_none() {
                    if let Ok(value) = input.try(|input| mask_image::single_value
                                                                   ::parse(context, input)) {
                        image = Some(value);
                        continue
                    }
                }
                if position.is_none() {
                    if let Ok(value) = input.try(|input| Position::parse(context, input)) {
                        position = Some(value);

                        // Parse mask size, if applicable.
                        size = input.try(|input| {
                            input.expect_delim('/')?;
                            mask_size::single_value::parse(context, input)
                        }).ok();

                        continue
                    }
                }
                    if repeat.is_none() {
                        if let Ok(value) = input.try(|input| mask_repeat::single_value
                                                                               ::parse(context, input)) {
                            repeat = Some(value);
                            continue
                        }
                    }
                    if origin.is_none() {
                        if let Ok(value) = input.try(|input| mask_origin::single_value
                                                                               ::parse(context, input)) {
                            origin = Some(value);
                            continue
                        }
                    }
                    if clip.is_none() {
                        if let Ok(value) = input.try(|input| mask_clip::single_value
                                                                               ::parse(context, input)) {
                            clip = Some(value);
                            continue
                        }
                    }
                    if composite.is_none() {
                        if let Ok(value) = input.try(|input| mask_composite::single_value
                                                                               ::parse(context, input)) {
                            composite = Some(value);
                            continue
                        }
                    }
                    if mode.is_none() {
                        if let Ok(value) = input.try(|input| mask_mode::single_value
                                                                               ::parse(context, input)) {
                            mode = Some(value);
                            continue
                        }
                    }
                break
            }
            if clip.is_none() {
                if let Some(origin) = origin {
                    clip = Some(mask_clip::single_value::SpecifiedValue::from(origin));
                }
            }
            let mut any = false;
                any = any || image.is_some();
                any = any || mode.is_some();
                any = any || position.is_some();
                any = any || size.is_some();
                any = any || repeat.is_some();
                any = any || origin.is_some();
                any = any || clip.is_some();
                any = any || composite.is_some();
            if any {
                if let Some(position) = position {
                    mask_position_x.push(position.horizontal);
                    mask_position_y.push(position.vertical);
                } else {
                    mask_position_x.push(PositionComponent::zero());
                    mask_position_y.push(PositionComponent::zero());
                }
                    if let Some(m_image) = image {
                        mask_image.push(m_image);
                    } else {
                        mask_image.push(mask_image::single_value
                                                        ::get_initial_specified_value());
                    }
                    if let Some(m_mode) = mode {
                        mask_mode.push(m_mode);
                    } else {
                        mask_mode.push(mask_mode::single_value
                                                        ::get_initial_specified_value());
                    }
                    if let Some(m_size) = size {
                        mask_size.push(m_size);
                    } else {
                        mask_size.push(mask_size::single_value
                                                        ::get_initial_specified_value());
                    }
                    if let Some(m_repeat) = repeat {
                        mask_repeat.push(m_repeat);
                    } else {
                        mask_repeat.push(mask_repeat::single_value
                                                        ::get_initial_specified_value());
                    }
                    if let Some(m_origin) = origin {
                        mask_origin.push(m_origin);
                    } else {
                        mask_origin.push(mask_origin::single_value
                                                        ::get_initial_specified_value());
                    }
                    if let Some(m_clip) = clip {
                        mask_clip.push(m_clip);
                    } else {
                        mask_clip.push(mask_clip::single_value
                                                        ::get_initial_specified_value());
                    }
                    if let Some(m_composite) = composite {
                        mask_composite.push(m_composite);
                    } else {
                        mask_composite.push(mask_composite::single_value
                                                        ::get_initial_specified_value());
                    }
                Ok(())
            } else {
                Err(input.new_custom_error(StyleParseErrorKind::UnspecifiedError))
            }
        })?;

        Ok(expanded! {
                mask_image: mask_image::SpecifiedValue(mask_image.into()),
                mask_mode: mask_mode::SpecifiedValue(mask_mode.into()),
                mask_position_x: mask_position_x::SpecifiedValue(mask_position_x.into()),
                mask_position_y: mask_position_y::SpecifiedValue(mask_position_y.into()),
                mask_size: mask_size::SpecifiedValue(mask_size.into()),
                mask_repeat: mask_repeat::SpecifiedValue(mask_repeat.into()),
                mask_origin: mask_origin::SpecifiedValue(mask_origin.into()),
                mask_clip: mask_clip::SpecifiedValue(mask_clip.into()),
                mask_composite: mask_composite::SpecifiedValue(mask_composite.into()),
         })
    }

    impl<'a> ToCss for LonghandsToSerialize<'a>  {
        fn to_css<W>(&self, dest: &mut CssWriter<W>) -> fmt::Result where W: fmt::Write {
            use crate::properties::longhands::mask_origin::single_value::computed_value::T as Origin;
            use crate::properties::longhands::mask_clip::single_value::computed_value::T as Clip;

            let len = self.mask_image.0.len();
            if len == 0 {
                return Ok(());
            }
                if self.mask_mode.0.len() != len {
                    return Ok(());
                }
                if self.mask_position_x.0.len() != len {
                    return Ok(());
                }
                if self.mask_position_y.0.len() != len {
                    return Ok(());
                }
                if self.mask_size.0.len() != len {
                    return Ok(());
                }
                if self.mask_repeat.0.len() != len {
                    return Ok(());
                }
                if self.mask_origin.0.len() != len {
                    return Ok(());
                }
                if self.mask_clip.0.len() != len {
                    return Ok(());
                }
                if self.mask_composite.0.len() != len {
                    return Ok(());
                }

            for i in 0..len {
                if i > 0 {
                    dest.write_str(", ")?;
                }

                    let image = &self.mask_image.0[i];
                    let mode = &self.mask_mode.0[i];
                    let position_x = &self.mask_position_x.0[i];
                    let position_y = &self.mask_position_y.0[i];
                    let size = &self.mask_size.0[i];
                    let repeat = &self.mask_repeat.0[i];
                    let origin = &self.mask_origin.0[i];
                    let clip = &self.mask_clip.0[i];
                    let composite = &self.mask_composite.0[i];

                image.to_css(dest)?;

                if *mode != mask_mode::single_value::get_initial_specified_value() {
                    dest.write_str(" ")?;
                    mode.to_css(dest)?;
                }

                if *position_x != PositionComponent::zero() ||
                    *position_y != PositionComponent::zero() ||
                    *size != mask_size::single_value::get_initial_specified_value()
                {
                    dest.write_str(" ")?;
                    Position {
                        horizontal: position_x.clone(),
                        vertical: position_y.clone()
                    }.to_css(dest)?;

                    if *size != mask_size::single_value::get_initial_specified_value() {
                        dest.write_str(" / ")?;
                        size.to_css(dest)?;
                    }
                }

                if *repeat != mask_repeat::single_value::get_initial_specified_value() {
                    dest.write_str(" ")?;
                    repeat.to_css(dest)?;
                }

                if *origin != Origin::BorderBox || *clip != Clip::BorderBox {
                    dest.write_str(" ")?;
                    origin.to_css(dest)?;
                    if *clip != From::from(*origin) {
                        dest.write_str(" ")?;
                        clip.to_css(dest)?;
                    }
                }

                if *composite != mask_composite::single_value::get_initial_specified_value() {
                    dest.write_str(" ")?;
                    composite.to_css(dest)?;
                }
            }

            Ok(())
        }
    }

    }




    /// https://drafts.csswg.org/css-masks-4/#the-mask-position
    pub mod mask_position {
        use cssparser::Parser;
        use crate::parser::ParserContext;
        use crate::properties::{PropertyDeclaration, SourcePropertyDeclaration, MaybeBoxed, longhands};
        #[allow(unused_imports)]
        use selectors::parser::SelectorParseErrorKind;
        #[allow(unused_imports)]
        use std::fmt::{self, Write};
        #[allow(unused_imports)]
        use style_traits::{ParseError, StyleParseErrorKind};
        #[allow(unused_imports)]
        use style_traits::{CssWriter, KeywordsCollectFn, SpecifiedValueInfo, ToCss};

        #[derive(SpecifiedValueInfo)]
        pub struct Longhands {
                pub mask_position_x:
                    longhands::mask_position_x::SpecifiedValue
                    ,
                pub mask_position_y:
                    longhands::mask_position_y::SpecifiedValue
                    ,
        }

        /// Represents a serializable set of all of the longhand properties that
        /// correspond to a shorthand.
        pub struct LonghandsToSerialize<'a> {
                pub mask_position_x:
                    &'a longhands::mask_position_x::SpecifiedValue,
                pub mask_position_y:
                    &'a longhands::mask_position_y::SpecifiedValue,
        }

        impl<'a> LonghandsToSerialize<'a> {
            /// Tries to get a serializable set of longhands given a set of
            /// property declarations.
            pub fn from_iter<I>(iter: I) -> Result<Self, ()>
            where
                I: Iterator<Item=&'a PropertyDeclaration>,
            {
                // Define all of the expected variables that correspond to the shorthand
                    let mut mask_position_x = None;
                    let mut mask_position_y = None;

                // Attempt to assign the incoming declarations to the expected variables
                for longhand in iter {
                    match *longhand {
                            PropertyDeclaration::MaskPositionX(ref value) => {
                                mask_position_x = Some(value)
                            },
                            PropertyDeclaration::MaskPositionY(ref value) => {
                                mask_position_y = Some(value)
                            },
                        _ => {}
                    };
                }

                // If any of the expected variables are missing, return an error
                match (
                        mask_position_x,
                        mask_position_y,
                ) {

                    (
                        Some(mask_position_x),
                        Some(mask_position_y),
                    ) =>
                    Ok(LonghandsToSerialize {
                            mask_position_x,
                            mask_position_y,
                    }),
                    _ => Err(())
                }
            }
        }

        /// Parse the given shorthand and fill the result into the
        /// `declarations` vector.
        pub fn parse_into<'i, 't>(
            declarations: &mut SourcePropertyDeclaration,
            context: &ParserContext,
            input: &mut Parser<'i, 't>,
        ) -> Result<(), ParseError<'i>> {
            #[allow(unused_imports)]
            use crate::properties::{NonCustomPropertyId, LonghandId};
            input.parse_entirely(|input| parse_value(context, input)).map(|longhands| {
                    declarations.push(PropertyDeclaration::MaskPositionX(
                        longhands.mask_position_x
                    ));
                    declarations.push(PropertyDeclaration::MaskPositionY(
                        longhands.mask_position_y
                    ));
            })
        }

        
    use crate::properties::longhands::{mask_position_x,mask_position_y};
    use crate::values::specified::position::Position;
    use crate::parser::Parse;

    pub fn parse_value<'i, 't>(
        context: &ParserContext,
        input: &mut Parser<'i, 't>,
    ) -> Result<Longhands, ParseError<'i>> {
        // Vec grows from 0 to 4 by default on first push().  So allocate with
        // capacity 1, so in the common case of only one item we don't way
        // overallocate, then shrink.  Note that we always push at least one
        // item if parsing succeeds.
        let mut position_x = Vec::with_capacity(1);
        let mut position_y = Vec::with_capacity(1);
        let mut any = false;

        input.parse_comma_separated(|input| {
            let value = Position::parse(context, input)?;
            position_x.push(value.horizontal);
            position_y.push(value.vertical);
            any = true;
            Ok(())
        })?;

        if !any {
            return Err(input.new_custom_error(StyleParseErrorKind::UnspecifiedError));
        }


        Ok(expanded! {
            mask_position_x: mask_position_x::SpecifiedValue(position_x.into()),
            mask_position_y: mask_position_y::SpecifiedValue(position_y.into()),
        })
    }

    impl<'a> ToCss for LonghandsToSerialize<'a>  {
        fn to_css<W>(&self, dest: &mut CssWriter<W>) -> fmt::Result where W: fmt::Write {
            let len = self.mask_position_x.0.len();
            if len == 0 || self.mask_position_y.0.len() != len {
                return Ok(());
            }

            for i in 0..len {
                Position {
                    horizontal: self.mask_position_x.0[i].clone(),
                    vertical: self.mask_position_y.0[i].clone()
                }.to_css(dest)?;

                if i < len - 1 {
                    dest.write_str(", ")?;
                }
            }

            Ok(())
        }
    }

    }

